import typing
from . import order_status
from .order_status import OrderStatusKind, OrderStatusJSON
from . import order_type
from .order_type import OrderTypeKind, OrderTypeJSON
from . import update_global_config_mode
from .update_global_config_mode import (
    UpdateGlobalConfigModeKind,
    UpdateGlobalConfigModeJSON,
)
from . import update_global_config_value
from .update_global_config_value import (
    UpdateGlobalConfigValueKind,
    UpdateGlobalConfigValueJSON,
)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class SetRouterSplitArgsJSON(typing.TypedDict):
    split_router: int


@dataclass
class SetRouterSplitArgs:
    layout: typing.ClassVar = borsh.CStruct("split_router" / borsh.U64)
    split_router: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "SetRouterSplitArgs":
        return cls(split_router=obj.split_router)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"split_router": self.split_router}

    def to_json(self) -> SetRouterSplitArgsJSON:
        return {"split_router": self.split_router}

    @classmethod
    def from_json(cls, obj: SetRouterSplitArgsJSON) -> "SetRouterSplitArgs":
        return cls(split_router=obj["split_router"])



from .initialize import initialize, InitializeArgs, InitializeAccounts
from .set_admin import set_admin, SetAdminAccounts
from .set_relayer import set_relayer, SetRelayerAccounts
from .set_splits import set_splits, SetSplitsArgs, SetSplitsAccounts
from .set_router_split import (
    set_router_split,
    SetRouterSplitArgs,
    SetRouterSplitAccounts,
)
from .submit_bid import submit_bid, SubmitBidArgs, SubmitBidAccounts
from .check_permission import check_permission, CheckPermissionAccounts
from .withdraw_fees import withdraw_fees, WithdrawFeesAccounts



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class CreateOrderArgs(typing.TypedDict):
    input_amount: int
    output_amount: int
    order_type: int


layout = borsh.CStruct(
    "input_amount" / borsh.U64, "output_amount" / borsh.U64, "order_type" / borsh.U8
)


class CreateOrderAccounts(typing.TypedDict):
    maker: Pubkey
    global_config: Pubkey
    pda_authority: Pubkey
    order: Pubkey
    input_mint: Pubkey
    output_mint: Pubkey
    maker_ata: Pubkey
    input_vault: Pubkey
    input_token_program: Pubkey
    output_token_program: Pubkey


def create_order(
    args: CreateOrderArgs,
    accounts: CreateOrderAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["maker"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["global_config"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["pda_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["order"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["input_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["output_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["maker_ata"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["input_vault"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["input_token_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["output_token_program"], is_signer=False, is_writable=False
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x8d6%\xcf\xed\xd2\xfa\xd7"
    encoded_args = layout.build(
        {
            "input_amount": args["input_amount"],
            "output_amount": args["output_amount"],
            "order_type": args["order_type"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from typing import Dict, TypedDict

from solders.pubkey import Pubkey

from express_relay.models import OpportunityAdapterConfig

OPPORTUNITY_ADAPTER_CONFIGS = {
    "op_sepolia": OpportunityAdapterConfig(
        chain_id=11155420,
        opportunity_adapter_factory="0xfA119693864b2F185742A409c66f04865c787754",
        opportunity_adapter_init_bytecode_hash="0x3d71516d94b96a8fdca4e3a5825a6b41c9268a8e94610367e69a8462cc543533",
        permit2="0x000000000022D473030F116dDEE9F6B43aC78BA3",
        weth="0x74A4A85C611679B73F402B36c0F84A7D2CcdFDa3",
    ),
    "mode": OpportunityAdapterConfig(
        chain_id=34443,
        opportunity_adapter_factory="0x59F78DE21a0b05d96Ae00c547BA951a3B905602f",
        opportunity_adapter_init_bytecode_hash="0xd53b8e32ab2ecba07c3e3a17c3c5e492c62e2f7051b89e5154f52e6bfeb0e38f",
        permit2="0x000000000022D473030F116dDEE9F6B43aC78BA3",
        weth="0x4200000000000000000000000000000000000006",
    ),
}

PERMIT_BATCH_TRANSFER_FROM_TYPESTRING = "((address,uint256)[],uint256,uint256)"
EXECUTION_WITNESS_TYPESTRING = (
    "((address,uint256)[],address,address,bytes,uint256,uint256)"
)
EXECUTION_PARAMS_TYPESTRING = (
    f"({PERMIT_BATCH_TRANSFER_FROM_TYPESTRING},{EXECUTION_WITNESS_TYPESTRING})"
)


class SvmProgramConfig(TypedDict):
    express_relay_program: Pubkey


SVM_CONFIGS: Dict[str, SvmProgramConfig] = {
    "development-solana": {
        "express_relay_program": Pubkey.from_string(
            "PytERJFhAKuNNuaiXkApLfWzwNwSNDACpigT3LwQfou"
        ),
    }
}



from .initialize_global_config import (
    initialize_global_config,
    InitializeGlobalConfigAccounts,
)
from .initialize_vault import initialize_vault, InitializeVaultAccounts
from .create_order import create_order, CreateOrderArgs, CreateOrderAccounts
from .close_order_and_claim_tip import (
    close_order_and_claim_tip,
    CloseOrderAndClaimTipAccounts,
)
from .take_order import take_order, TakeOrderArgs, TakeOrderAccounts
from .flash_take_order_start import (
    flash_take_order_start,
    FlashTakeOrderStartArgs,
    FlashTakeOrderStartAccounts,
)
from .flash_take_order_end import (
    flash_take_order_end,
    FlashTakeOrderEndArgs,
    FlashTakeOrderEndAccounts,
)
from .update_global_config import (
    update_global_config,
    UpdateGlobalConfigArgs,
    UpdateGlobalConfigAccounts,
)
from .withdraw_host_tip import withdraw_host_tip, WithdrawHostTipAccounts



import base64
from datetime import datetime
from enum import Enum
from typing import Any, Annotated, ClassVar

from pydantic import (
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    BaseModel,
    model_validator,
    Field,
    Base64Bytes,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema
from solders.hash import Hash as _SvmHash
from solders.pubkey import Pubkey as _SvmAddress
from solders.signature import Signature as _SvmSignature
from solders.transaction import Transaction as _SvmTransaction

from express_relay.svm.generated.limo.accounts.order import Order
from express_relay.models.base import (
    IntString,
    UUIDString,
    UnsupportedOpportunityDeleteVersionException,
    UnsupportedOpportunityVersionException,
    BidStatusVariantsSvm,
)


class _TransactionPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        def validate_from_str(value: str) -> _SvmTransaction:
            return _SvmTransaction.from_bytes(base64.b64decode(value))

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(_SvmTransaction),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: base64.b64encode(bytes(instance)).decode("utf-8")
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


class _SvmAddressPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        def validate_from_str(value: str) -> _SvmAddress:
            return _SvmAddress.from_string(value)

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(_SvmTransaction),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: str(instance)
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


class _HashPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        def validate_from_str(value: str) -> _SvmHash:
            return _SvmHash.from_string(value)

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(Order),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


class _SignaturePydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        def validate_from_str(value: str) -> _SvmSignature:
            return _SvmSignature.from_string(value)

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(Order),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


SvmTransaction = Annotated[_SvmTransaction, _TransactionPydanticAnnotation]
SvmAddress = Annotated[_SvmAddress, _SvmAddressPydanticAnnotation]
SvmHash = Annotated[_SvmHash, _HashPydanticAnnotation]
SvmSignature = Annotated[_SvmSignature, _SignaturePydanticAnnotation]


class BidSvm(BaseModel):
    """
    Attributes:
        transaction: The transaction including the bid
        chain_id: The chain ID to bid on.
    """

    transaction: SvmTransaction
    chain_id: str


class _OrderPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        def validate_from_str(value: str) -> Order:
            return Order.decode(base64.b64decode(value))

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(Order),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: base64.b64encode(Order.layout.build(instance)).decode(
                    "utf-8"
                )
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


class OpportunitySvm(BaseModel):
    """
    Attributes:
        chain_id: The chain ID to bid on.
        version: The version of the opportunity.
        creation_time: The creation time of the opportunity.
        opportunity_id: The ID of the opportunity.
        slot: The slot where this order was created or updated
        program: The program which handles this opportunity
        order: The order to be executed.
        order_address: The address of the order.
    """

    chain_id: str
    version: str
    creation_time: IntString
    opportunity_id: UUIDString

    slot: int

    program: str
    order: Annotated[Order, _OrderPydanticAnnotation]
    order_address: SvmAddress

    supported_versions: ClassVar[list[str]] = ["v1"]
    supported_programs: ClassVar[list[str]] = ["limo"]

    @model_validator(mode="before")
    @classmethod
    def check_version(cls, data):
        if data["version"] not in cls.supported_versions:
            raise UnsupportedOpportunityVersionException(
                f"Cannot handle opportunity version: {data['version']}. Please upgrade your client."
            )
        return data


class BidStatusSvm(BaseModel):
    """
    Attributes:
        type: The current status of the bid.
        result: The result of the bid: a transaction hash if the status is SUBMITTED or WON.
                The LOST status may have a result.
    """

    type: BidStatusVariantsSvm
    result: SvmSignature | None = Field(default=None)

    @model_validator(mode="after")
    def check_result(self):
        if self.type == BidStatusVariantsSvm.WON or self.type == BidStatusVariantsSvm.SUBMITTED:
            assert (
                self.result is not None
            ), "bid result should not be empty when status is won or submitted"
        return self


class BidResponseSvm(BaseModel):
    """
    Attributes:
        id: The unique id for bid.
        bid_amount: The amount of the bid in lamports.
        chain_id: The chain ID to bid on.
        permission_key: The permission key to bid on.
        status: The latest status for bid.
        initiation_time: The time server received the bid formatted in rfc3339.
        profile_id: The profile id for the bid owner.
    """

    id: UUIDString
    bid_amount: int
    chain_id: str
    permission_key: Base64Bytes
    status: BidStatusSvm
    initiation_time: datetime
    transaction: SvmTransaction
    profile_id: str | None = Field(default=None)

class SvmChainUpdate(BaseModel):
    """
    Attributes:
        chain_id: The chain ID corresponding to the update.
        blockhash: A recent blockhash from the chain.
    """
    chain_id: str
    blockhash: SvmHash


class ProgramSvm(Enum):
    LIMO = "limo"
    PHANTOM = "phantom"


class OpportunityDeleteSvm(BaseModel):
    """
    Attributes:
        chain_id: The chain ID for opportunities to be removed.
        program: The program which this opportunities to be removed.
        permission_account: The permission account for the opportunities to be removed.
        router: The router for opportunties to be removed.
    """
    chain_id: str
    program: ProgramSvm
    permission_account: SvmAddress
    router: SvmAddress
    version: str

    supported_versions: ClassVar[list[str]] = ["v1"]
    chain_type: ClassVar[str] = "svm"

    @model_validator(mode="before")
    @classmethod
    def check_version(cls, data):
        if data["version"] not in cls.supported_versions:
            raise UnsupportedOpportunityDeleteVersionException(
                f"Cannot handle opportunity version: {data['version']}. Please upgrade your client."
            )
        return data



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class ActiveJSON(typing.TypedDict):
    kind: typing.Literal["Active"]


class FilledJSON(typing.TypedDict):
    kind: typing.Literal["Filled"]


class CancelledJSON(typing.TypedDict):
    kind: typing.Literal["Cancelled"]


@dataclass
class Active:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Active"

    @classmethod
    def to_json(cls) -> ActiveJSON:
        return ActiveJSON(
            kind="Active",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Active": {},
        }


@dataclass
class Filled:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Filled"

    @classmethod
    def to_json(cls) -> FilledJSON:
        return FilledJSON(
            kind="Filled",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Filled": {},
        }


@dataclass
class Cancelled:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Cancelled"

    @classmethod
    def to_json(cls) -> CancelledJSON:
        return CancelledJSON(
            kind="Cancelled",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Cancelled": {},
        }


OrderStatusKind = typing.Union[Active, Filled, Cancelled]
OrderStatusJSON = typing.Union[ActiveJSON, FilledJSON, CancelledJSON]


def from_decoded(obj: dict) -> OrderStatusKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Active" in obj:
        return Active()
    if "Filled" in obj:
        return Filled()
    if "Cancelled" in obj:
        return Cancelled()
    raise ValueError("Invalid enum object")


def from_json(obj: OrderStatusJSON) -> OrderStatusKind:
    if obj["kind"] == "Active":
        return Active()
    if obj["kind"] == "Filled":
        return Filled()
    if obj["kind"] == "Cancelled":
        return Cancelled()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Active" / borsh.CStruct(),
    "Filled" / borsh.CStruct(),
    "Cancelled" / borsh.CStruct(),
)



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class OrderJSON(typing.TypedDict):
    global_config: str
    maker: str
    input_mint: str
    input_mint_program_id: str
    output_mint: str
    output_mint_program_id: str
    initial_input_amount: int
    expected_output_amount: int
    remaining_input_amount: int
    filled_output_amount: int
    tip_amount: int
    number_of_fills: int
    order_type: int
    status: int
    in_vault_bump: int
    flash_ix_lock: int
    padding0: list[int]
    last_updated_timestamp: int
    padding: list[int]


@dataclass
class Order:
    discriminator: typing.ClassVar = b"\x86\xad\xdf\xb9MV\x1c3"
    layout: typing.ClassVar = borsh.CStruct(
        "global_config" / BorshPubkey,
        "maker" / BorshPubkey,
        "input_mint" / BorshPubkey,
        "input_mint_program_id" / BorshPubkey,
        "output_mint" / BorshPubkey,
        "output_mint_program_id" / BorshPubkey,
        "initial_input_amount" / borsh.U64,
        "expected_output_amount" / borsh.U64,
        "remaining_input_amount" / borsh.U64,
        "filled_output_amount" / borsh.U64,
        "tip_amount" / borsh.U64,
        "number_of_fills" / borsh.U64,
        "order_type" / borsh.U8,
        "status" / borsh.U8,
        "in_vault_bump" / borsh.U8,
        "flash_ix_lock" / borsh.U8,
        "padding0" / borsh.U8[4],
        "last_updated_timestamp" / borsh.U64,
        "padding" / borsh.U64[20],
    )
    global_config: Pubkey
    maker: Pubkey
    input_mint: Pubkey
    input_mint_program_id: Pubkey
    output_mint: Pubkey
    output_mint_program_id: Pubkey
    initial_input_amount: int
    expected_output_amount: int
    remaining_input_amount: int
    filled_output_amount: int
    tip_amount: int
    number_of_fills: int
    order_type: int
    status: int
    in_vault_bump: int
    flash_ix_lock: int
    padding0: list[int]
    last_updated_timestamp: int
    padding: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Order"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Order"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Order"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Order":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Order.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            global_config=dec.global_config,
            maker=dec.maker,
            input_mint=dec.input_mint,
            input_mint_program_id=dec.input_mint_program_id,
            output_mint=dec.output_mint,
            output_mint_program_id=dec.output_mint_program_id,
            initial_input_amount=dec.initial_input_amount,
            expected_output_amount=dec.expected_output_amount,
            remaining_input_amount=dec.remaining_input_amount,
            filled_output_amount=dec.filled_output_amount,
            tip_amount=dec.tip_amount,
            number_of_fills=dec.number_of_fills,
            order_type=dec.order_type,
            status=dec.status,
            in_vault_bump=dec.in_vault_bump,
            flash_ix_lock=dec.flash_ix_lock,
            padding0=dec.padding0,
            last_updated_timestamp=dec.last_updated_timestamp,
            padding=dec.padding,
        )

    def to_json(self) -> OrderJSON:
        return {
            "global_config": str(self.global_config),
            "maker": str(self.maker),
            "input_mint": str(self.input_mint),
            "input_mint_program_id": str(self.input_mint_program_id),
            "output_mint": str(self.output_mint),
            "output_mint_program_id": str(self.output_mint_program_id),
            "initial_input_amount": self.initial_input_amount,
            "expected_output_amount": self.expected_output_amount,
            "remaining_input_amount": self.remaining_input_amount,
            "filled_output_amount": self.filled_output_amount,
            "tip_amount": self.tip_amount,
            "number_of_fills": self.number_of_fills,
            "order_type": self.order_type,
            "status": self.status,
            "in_vault_bump": self.in_vault_bump,
            "flash_ix_lock": self.flash_ix_lock,
            "padding0": self.padding0,
            "last_updated_timestamp": self.last_updated_timestamp,
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: OrderJSON) -> "Order":
        return cls(
            global_config=Pubkey.from_string(obj["global_config"]),
            maker=Pubkey.from_string(obj["maker"]),
            input_mint=Pubkey.from_string(obj["input_mint"]),
            input_mint_program_id=Pubkey.from_string(obj["input_mint_program_id"]),
            output_mint=Pubkey.from_string(obj["output_mint"]),
            output_mint_program_id=Pubkey.from_string(obj["output_mint_program_id"]),
            initial_input_amount=obj["initial_input_amount"],
            expected_output_amount=obj["expected_output_amount"],
            remaining_input_amount=obj["remaining_input_amount"],
            filled_output_amount=obj["filled_output_amount"],
            tip_amount=obj["tip_amount"],
            number_of_fills=obj["number_of_fills"],
            order_type=obj["order_type"],
            status=obj["status"],
            in_vault_bump=obj["in_vault_bump"],
            flash_ix_lock=obj["flash_ix_lock"],
            padding0=obj["padding0"],
            last_updated_timestamp=obj["last_updated_timestamp"],
            padding=obj["padding"],
        )



import argparse
import asyncio
import logging
import typing
from decimal import Decimal
from typing import List

from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Finalized
from solders.instruction import Instruction
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solders.transaction import Transaction

from express_relay.client import (
    ExpressRelayClient,
)
from express_relay.constants import SVM_CONFIGS
from express_relay.models import BidStatusUpdate, Opportunity, OpportunityDelete
from express_relay.models.base import BidStatusVariantsSvm
from express_relay.models.svm import BidSvm, OpportunitySvm, SvmChainUpdate, SvmHash
from express_relay.svm.generated.express_relay.accounts.express_relay_metadata import (
    ExpressRelayMetadata,
)
from express_relay.svm.generated.express_relay.program_id import (
    PROGRAM_ID as SVM_EXPRESS_RELAY_PROGRAM_ID,
)
from express_relay.svm.limo_client import LimoClient, OrderStateAndAddress

DEADLINE = 2**62


class SimpleSearcherSvm:
    express_relay_metadata: ExpressRelayMetadata | None
    mint_decimals_cache: typing.Dict[str, int]
    recent_blockhash: typing.Dict[str, SvmHash]

    def __init__(
        self,
        server_url: str,
        private_key: Keypair,
        bid_amount: int,
        chain_id: str,
        svm_rpc_endpoint: str,
        api_key: str | None = None,
    ):
        self.client = ExpressRelayClient(
            server_url,
            api_key,
            self.opportunity_callback,
            self.bid_status_callback,
            self.svm_chain_update_callback,
            self.remove_opportunities_callback,
        )
        self.private_key = private_key
        self.bid_amount = bid_amount
        self.chain_id = chain_id
        if self.chain_id not in SVM_CONFIGS:
            raise ValueError(f"Chain ID {self.chain_id} not supported")
        self.svm_config = SVM_CONFIGS[self.chain_id]
        self.rpc_client = AsyncClient(svm_rpc_endpoint)
        self.limo_client = LimoClient(self.rpc_client)
        self.express_relay_metadata = None
        self.mint_decimals_cache = {}
        self.recent_blockhash = {}

        self.logger = logging.getLogger('searcher')
        self.setup_logger()
        self.logger.info("Using searcher pubkey: %s", self.private_key.pubkey())

    def setup_logger(self):
        self.logger.setLevel(logging.INFO)
        log_handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s %(levelname)s:%(name)s:%(module)s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        log_handler.setFormatter(formatter)
        self.logger.addHandler(log_handler)

    async def opportunity_callback(self, opp: Opportunity):
        """
        Callback function to run when a new opportunity is found.

        Args:
            opp: An object representing a single opportunity.
        """
        if opp.chain_id not in self.recent_blockhash:
            self.logger.info(f"No recent blockhash for chain, {opp.chain_id} skipping bid")
            return None

        bid = await self.assess_opportunity(typing.cast(OpportunitySvm, opp))

        if bid:
            try:
                bid_id = await self.client.submit_bid(bid)
                self.logger.info(f"Submitted bid {str(bid_id)} for opportunity {str(opp.opportunity_id)}")
            except Exception as e:
                self.logger.error(
                    f"Error submitting bid for opportunity {str(opp.opportunity_id)}: {e}"
                )

    async def bid_status_callback(self, bid_status_update: BidStatusUpdate):
        """
        Callback function to run when a bid status is updated.

        Args:
            bid_status_update: An object representing an update to the status of a bid.
        """
        id = bid_status_update.id
        status = bid_status_update.bid_status.type
        result = bid_status_update.bid_status.result

        result_details = ""
        if status == BidStatusVariantsSvm.SUBMITTED or status == BidStatusVariantsSvm.WON:
            result_details = f", transaction {result}"
        elif status == BidStatusVariantsSvm.LOST:
            if result:
                result_details = f", transaction {result}"
        self.logger.info(f"Bid status for bid {id}: {status.value}{result_details}")

    async def get_mint_decimals(self, mint: Pubkey) -> int:
        if str(mint) not in self.mint_decimals_cache:
            self.mint_decimals_cache[
                str(mint)
            ] = await self.limo_client.get_mint_decimals(mint)
        return self.mint_decimals_cache[str(mint)]

    async def assess_opportunity(self, opp: OpportunitySvm) -> BidSvm | None:
        """
        Method to assess an opportunity and return a bid if the opportunity is worth taking. This method always returns a bid for any valid opportunity. The transaction in this bid transfers assets from the searcher's wallet to fulfill the limit order.

        Args:
            opp: An object representing a single opportunity.
        Returns:
            A bid object if the opportunity is worth taking to be submitted to the Express Relay server, otherwise None.
        """
        order: OrderStateAndAddress = {"address": opp.order_address, "state": opp.order}

        ixs_take_order = await self.generate_take_order_ixs(order)
        bid_amount = await self.get_bid_amount(order)
        router = self.limo_client.get_pda_authority(
            self.limo_client.get_program_id(), order["state"].global_config
        )

        submit_bid_ix = self.client.get_svm_submit_bid_instruction(
            searcher=self.private_key.pubkey(),
            router=router,
            permission_key=order["address"],
            bid_amount=bid_amount,
            deadline=DEADLINE,
            chain_id=self.chain_id,
            fee_receiver_relayer=(await self.get_metadata()).fee_receiver_relayer,
            relayer_signer=(await self.get_metadata()).relayer_signer,
        )
        transaction = Transaction.new_with_payer(
            [submit_bid_ix] + ixs_take_order, self.private_key.pubkey()
        )
        transaction.partial_sign(
            [self.private_key], recent_blockhash=self.recent_blockhash[self.chain_id]
        )
        bid = BidSvm(transaction=transaction, chain_id=self.chain_id)
        return bid

    async def generate_take_order_ixs(
        self, order: OrderStateAndAddress
    ) -> List[Instruction]:
        """
        Helper method to form the Limo instructions to take an order.

        Args:
            order: An object representing the order to be fulfilled.
        Returns:
            A list of Limo instructions to take an order.
        """
        input_amount = self.get_input_amount(order)
        output_amount = (
            order["state"].expected_output_amount * input_amount
            + order["state"].initial_input_amount
            - 1  # take the ceiling of the divison by adding order[state].initial_input_amount - 1
        ) // order["state"].initial_input_amount

        input_mint_decimals = await self.get_mint_decimals(order["state"].input_mint)
        output_mint_decimals = await self.get_mint_decimals(order["state"].output_mint)
        self.logger.info(
            f"Order address {order['address']}\n"
            f"Fill rate {input_amount / order['state'].initial_input_amount}\n"
            f"Sell token {order['state'].input_mint} amount: {Decimal(input_amount) / Decimal(10 ** input_mint_decimals)}\n"
            f"Buy token {order['state'].output_mint} amount: {Decimal(output_amount) / Decimal(10**output_mint_decimals)}"
        )
        ixs_take_order = await self.limo_client.take_order_ix(
            self.private_key.pubkey(),
            order,
            input_amount,
            output_amount,
            self.svm_config["express_relay_program"],
        )
        return ixs_take_order

    def get_input_amount(self, order: OrderStateAndAddress) -> int:
        return order["state"].remaining_input_amount

    async def get_metadata(self) -> ExpressRelayMetadata:
        if self.express_relay_metadata is None:
            self.express_relay_metadata = await ExpressRelayMetadata.fetch(
                self.rpc_client,
                self.limo_client.get_express_relay_metadata_pda(
                    SVM_EXPRESS_RELAY_PROGRAM_ID
                ),
                commitment=Finalized,
            )
            if self.express_relay_metadata is None:
                raise ValueError("Express relay metadata account not found")
        return self.express_relay_metadata

    async def get_bid_amount(self, order: OrderStateAndAddress) -> int:
        """
        Args:
            order: An object representing the order to be fulfilled.
        Returns:
            The amount of bid to submit for the opportunity in lamports.
        """

        return self.bid_amount

    async def svm_chain_update_callback(self, svm_chain_update: SvmChainUpdate):
        self.recent_blockhash[svm_chain_update.chain_id] = svm_chain_update.blockhash

    # NOTE: Developers are responsible for implementing custom removal logic specific to their use case.
    async def remove_opportunities_callback(
        self, opportunity_delete: OpportunityDelete
    ):
        print(f"Opportunities {opportunity_delete} don't exist anymore")


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--private-key",
        type=str,
        help="Private key of the searcher in base58 format",
    )
    group.add_argument(
        "--private-key-json-file",
        type=str,
        help="Path to a json file containing the private key of the searcher in array of bytes format",
    )
    parser.add_argument(
        "--chain-id",
        type=str,
        required=True,
        choices=SVM_CONFIGS.keys(),
        help="Chain ID of the SVM network to submit bids",
    )
    parser.add_argument(
        "--endpoint-express-relay",
        type=str,
        required=True,
        help="Server endpoint to use for submitting bids",
    )
    parser.add_argument(
        "--endpoint-svm",
        type=str,
        required=True,
        help="Server endpoint to use for submitting bids",
    )
    parser.add_argument(
        "--api-key",
        type=str,
        required=False,
        help="The API key of the searcher to authenticate with the server for fetching and submitting bids",
    )
    parser.add_argument(
        "--bid",
        type=int,
        default=100,
        help="The amount of bid to submit for each opportunity",
    )
    return parser


async def main():
    parser = get_parser()
    args = parser.parse_args()

    if args.private_key:
        searcher_keypair = Keypair.from_base58_string(args.private_key)
    else:
        with open(args.private_key_json_file, "r") as f:
            searcher_keypair = Keypair.from_json(f.read())

    searcher = SimpleSearcherSvm(
        args.endpoint_express_relay,
        searcher_keypair,
        args.bid,
        args.chain_id,
        args.endpoint_svm,
        args.api_key,
    )

    await searcher.client.subscribe_chains([args.chain_id])

    task = await searcher.client.get_ws_loop()
    await task


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class InitializeGlobalConfigAccounts(typing.TypedDict):
    admin_authority: Pubkey
    pda_authority: Pubkey
    global_config: Pubkey


def initialize_global_config(
    accounts: InitializeGlobalConfigAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["admin_authority"], is_signer=True, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["pda_authority"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["global_config"], is_signer=False, is_writable=True
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"q\xd8z\x83\xe1\xd1\x167"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import asyncio
import random
from decimal import Decimal

from solders.keypair import Keypair

from express_relay.models import (
    Opportunity
)
from express_relay.svm.limo_client import OrderStateAndAddress
from .simple_searcher_svm import get_parser, SimpleSearcherSvm


class TestingSearcherSvm(SimpleSearcherSvm):

    def __init__(self, server_url: str, private_key: Keypair, bid_amount: int, chain_id: str, svm_rpc_endpoint: str,
                 fill_rate: int, with_latency: bool, bid_margin: int, api_key: str | None = None):
        super().__init__(server_url, private_key, bid_amount, chain_id, svm_rpc_endpoint, api_key)
        self.fill_rate = fill_rate
        self.with_latency = with_latency
        self.bid_margin = bid_margin

    async def opportunity_callback(self, opp: Opportunity):
        if self.with_latency:
            latency = 0.5 * random.random()
            self.logger.info(f"Adding latency of {latency * 100}ms")
            await asyncio.sleep(latency)
        return await super().opportunity_callback(opp)

    async def get_bid_amount(self, order: OrderStateAndAddress) -> int:
        return self.bid_amount + random.randint(-self.bid_margin, self.bid_margin)

    def get_input_amount(self, order: OrderStateAndAddress) -> int:
        return min(super().get_input_amount(order), order["state"].initial_input_amount * self.fill_rate // 100)


async def main():
    parser = get_parser()
    parser.add_argument(
        "--fill-rate",
        type=int,
        default=100,
        help="How much of the initial order size to fill in percentage. Default is 100%",
    )
    parser.add_argument(
        "--with-latency",
        required=False,
        default=False,
        action="store_true",
        help="Whether to add random latency to the bid submission. Default is false",
    )
    parser.add_argument(
        "--bid-margin",
        required=False,
        type=int,
        default=0,
        help="The margin to add or subtract from the bid. For example, 1 means the bid range is [bid - 1, bid + 1]. Default is 0",
    )
    args = parser.parse_args()

    if args.private_key:
        searcher_keypair = Keypair.from_base58_string(args.private_key)
    else:
        with open(args.private_key_json_file, "r") as private_key_file:
            searcher_keypair = Keypair.from_json(private_key_file.read())

    searcher = TestingSearcherSvm(
        args.endpoint_express_relay,
        searcher_keypair,
        args.bid,
        args.chain_id,
        args.endpoint_svm,
        args.fill_rate,
        args.with_latency,
        args.bid_margin,
        args.api_key,
    )

    await searcher.client.subscribe_chains([args.chain_id])

    task = await searcher.client.get_ws_loop()
    await task


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class UpdateEmergencyModeJSON(typing.TypedDict):
    kind: typing.Literal["UpdateEmergencyMode"]


class UpdateFlashTakeOrderBlockedJSON(typing.TypedDict):
    kind: typing.Literal["UpdateFlashTakeOrderBlocked"]


class UpdateBlockNewOrdersJSON(typing.TypedDict):
    kind: typing.Literal["UpdateBlockNewOrders"]


class UpdateBlockOrderTakingJSON(typing.TypedDict):
    kind: typing.Literal["UpdateBlockOrderTaking"]


class UpdateHostFeeBpsJSON(typing.TypedDict):
    kind: typing.Literal["UpdateHostFeeBps"]


class UpdateAdminAuthorityJSON(typing.TypedDict):
    kind: typing.Literal["UpdateAdminAuthority"]


class UpdateOrderTakingPermissionlessJSON(typing.TypedDict):
    kind: typing.Literal["UpdateOrderTakingPermissionless"]


@dataclass
class UpdateEmergencyMode:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "UpdateEmergencyMode"

    @classmethod
    def to_json(cls) -> UpdateEmergencyModeJSON:
        return UpdateEmergencyModeJSON(
            kind="UpdateEmergencyMode",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UpdateEmergencyMode": {},
        }


@dataclass
class UpdateFlashTakeOrderBlocked:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "UpdateFlashTakeOrderBlocked"

    @classmethod
    def to_json(cls) -> UpdateFlashTakeOrderBlockedJSON:
        return UpdateFlashTakeOrderBlockedJSON(
            kind="UpdateFlashTakeOrderBlocked",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UpdateFlashTakeOrderBlocked": {},
        }


@dataclass
class UpdateBlockNewOrders:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "UpdateBlockNewOrders"

    @classmethod
    def to_json(cls) -> UpdateBlockNewOrdersJSON:
        return UpdateBlockNewOrdersJSON(
            kind="UpdateBlockNewOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UpdateBlockNewOrders": {},
        }


@dataclass
class UpdateBlockOrderTaking:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "UpdateBlockOrderTaking"

    @classmethod
    def to_json(cls) -> UpdateBlockOrderTakingJSON:
        return UpdateBlockOrderTakingJSON(
            kind="UpdateBlockOrderTaking",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UpdateBlockOrderTaking": {},
        }


@dataclass
class UpdateHostFeeBps:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "UpdateHostFeeBps"

    @classmethod
    def to_json(cls) -> UpdateHostFeeBpsJSON:
        return UpdateHostFeeBpsJSON(
            kind="UpdateHostFeeBps",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UpdateHostFeeBps": {},
        }


@dataclass
class UpdateAdminAuthority:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "UpdateAdminAuthority"

    @classmethod
    def to_json(cls) -> UpdateAdminAuthorityJSON:
        return UpdateAdminAuthorityJSON(
            kind="UpdateAdminAuthority",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UpdateAdminAuthority": {},
        }


@dataclass
class UpdateOrderTakingPermissionless:
    discriminator: typing.ClassVar = 6
    kind: typing.ClassVar = "UpdateOrderTakingPermissionless"

    @classmethod
    def to_json(cls) -> UpdateOrderTakingPermissionlessJSON:
        return UpdateOrderTakingPermissionlessJSON(
            kind="UpdateOrderTakingPermissionless",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UpdateOrderTakingPermissionless": {},
        }


UpdateGlobalConfigModeKind = typing.Union[
    UpdateEmergencyMode,
    UpdateFlashTakeOrderBlocked,
    UpdateBlockNewOrders,
    UpdateBlockOrderTaking,
    UpdateHostFeeBps,
    UpdateAdminAuthority,
    UpdateOrderTakingPermissionless,
]
UpdateGlobalConfigModeJSON = typing.Union[
    UpdateEmergencyModeJSON,
    UpdateFlashTakeOrderBlockedJSON,
    UpdateBlockNewOrdersJSON,
    UpdateBlockOrderTakingJSON,
    UpdateHostFeeBpsJSON,
    UpdateAdminAuthorityJSON,
    UpdateOrderTakingPermissionlessJSON,
]


def from_decoded(obj: dict) -> UpdateGlobalConfigModeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "UpdateEmergencyMode" in obj:
        return UpdateEmergencyMode()
    if "UpdateFlashTakeOrderBlocked" in obj:
        return UpdateFlashTakeOrderBlocked()
    if "UpdateBlockNewOrders" in obj:
        return UpdateBlockNewOrders()
    if "UpdateBlockOrderTaking" in obj:
        return UpdateBlockOrderTaking()
    if "UpdateHostFeeBps" in obj:
        return UpdateHostFeeBps()
    if "UpdateAdminAuthority" in obj:
        return UpdateAdminAuthority()
    if "UpdateOrderTakingPermissionless" in obj:
        return UpdateOrderTakingPermissionless()
    raise ValueError("Invalid enum object")


def from_json(obj: UpdateGlobalConfigModeJSON) -> UpdateGlobalConfigModeKind:
    if obj["kind"] == "UpdateEmergencyMode":
        return UpdateEmergencyMode()
    if obj["kind"] == "UpdateFlashTakeOrderBlocked":
        return UpdateFlashTakeOrderBlocked()
    if obj["kind"] == "UpdateBlockNewOrders":
        return UpdateBlockNewOrders()
    if obj["kind"] == "UpdateBlockOrderTaking":
        return UpdateBlockOrderTaking()
    if obj["kind"] == "UpdateHostFeeBps":
        return UpdateHostFeeBps()
    if obj["kind"] == "UpdateAdminAuthority":
        return UpdateAdminAuthority()
    if obj["kind"] == "UpdateOrderTakingPermissionless":
        return UpdateOrderTakingPermissionless()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "UpdateEmergencyMode" / borsh.CStruct(),
    "UpdateFlashTakeOrderBlocked" / borsh.CStruct(),
    "UpdateBlockNewOrders" / borsh.CStruct(),
    "UpdateBlockOrderTaking" / borsh.CStruct(),
    "UpdateHostFeeBps" / borsh.CStruct(),
    "UpdateAdminAuthority" / borsh.CStruct(),
    "UpdateOrderTakingPermissionless" / borsh.CStruct(),
)



import string
from datetime import datetime
from typing import ClassVar

import web3
from eth_account.datastructures import SignedMessage
from pydantic import BaseModel, model_validator, Field
from pydantic.functional_serializers import PlainSerializer
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

from express_relay.models.base import (
    IntString,
    UUIDString,
    UnsupportedOpportunityDeleteVersionException,
    UnsupportedOpportunityVersionException,
    BidStatusVariantsEvm,
)


def check_hex_string(s: str):
    """
    Validates that a string is a valid hex string.

    Args:
        s: The string to validate as a hex string. Can be '0x'-prefixed.
    """
    ind = 0
    if s.startswith("0x"):
        ind = 2

    assert all(
        c in string.hexdigits for c in s[ind:]
    ), "string is not a valid hex string"

    return s


def check_bytes32(s: str):
    """
    Validates that a string is a valid 32-byte hex string.

    Args:
        s: The string to validate as a 32-byte hex string. Can be '0x'-prefixed.
    """
    check_hex_string(s)
    ind = 0
    if s.startswith("0x"):
        ind = 2

    assert len(s[ind:]) == 64, "hex string is not 32 bytes long"

    return s


def check_address(s: str):
    """
    Validates that a string is a valid Ethereum address.

    Args:
        s: The string to validate as an Ethereum address. Can be '0x'-prefixed.
    """
    assert web3.Web3.is_address(s), "string is not a valid Ethereum address"
    return s


HexString = Annotated[str, AfterValidator(check_hex_string)]
Bytes32 = Annotated[str, AfterValidator(check_bytes32)]
Address = Annotated[str, AfterValidator(check_address)]

SignedMessageString = Annotated[
    SignedMessage, PlainSerializer(lambda x: bytes(x.signature).hex(), return_type=str)
]


class TokenAmount(BaseModel):
    """
    Attributes:
        token: The address of the token contract.
        amount: The amount of the token.
    """

    token: Address
    amount: IntString


class BidEvm(BaseModel):
    """
    Attributes:
        amount: The amount of the bid in wei.
        target_calldata: The calldata for the contract call.
        chain_id: The chain ID to bid on.
        target_contract: The contract address to call.
        permission_key: The permission key to bid on.
    """

    amount: IntString
    target_calldata: HexString
    chain_id: str
    target_contract: Address
    permission_key: HexString


class OpportunityEvm(BaseModel):
    """
    Attributes:
        target_calldata: The calldata for the contract call.
        chain_id: The chain ID to bid on.
        target_contract: The contract address to call.
        permission_key: The permission key to bid on.
        buy_tokens: The tokens to receive in the opportunity.
        sell_tokens: The tokens to spend in the opportunity.
        target_call_value: The value to send with the contract call.
        version: The version of the opportunity.
        creation_time: The creation time of the opportunity.
        opportunity_id: The ID of the opportunity.
    """

    target_calldata: HexString
    chain_id: str
    target_contract: Address
    permission_key: HexString
    buy_tokens: list[TokenAmount]
    sell_tokens: list[TokenAmount]
    target_call_value: IntString
    version: str
    creation_time: IntString
    opportunity_id: UUIDString

    supported_versions: ClassVar[list[str]] = ["v1"]

    @model_validator(mode="before")
    @classmethod
    def check_version(cls, data):
        if data["version"] not in cls.supported_versions:
            raise UnsupportedOpportunityVersionException(
                f"Cannot handle opportunity version: {data['version']}. Please upgrade your client."
            )
        return data


class BidStatusEvm(BaseModel):
    """
    Attributes:
        type: The current status of the bid.
        result: The result of the bid: a transaction hash if the status is SUBMITTED or WON.
                The LOST status may have a result.
        index: The index of the bid in the submitted transaction.
    """

    type: BidStatusVariantsEvm
    result: Bytes32 | None = Field(default=None)
    index: int | None = Field(default=None)

    @model_validator(mode="after")
    def check_result(self):
        if self.type == BidStatusVariantsEvm.PENDING:
            assert self.result is None, "result must be None"
        elif self.type == BidStatusVariantsEvm.LOST:
            pass
        else:
            assert self.result is not None, "result must be a valid 32-byte hash"
        return self

    @model_validator(mode="after")
    def check_index(self):
        if self.type == BidStatusVariantsEvm.SUBMITTED or self.type == BidStatusVariantsEvm.WON:
            assert self.index is not None, "index must be a valid integer"
        elif self.type == BidStatusVariantsEvm.LOST:
            pass
        else:
            assert self.index is None, "index must be None"
        return self


class BidResponseEvm(BaseModel):
    """
    Attributes:
        id: The unique id for bid.
        amount: The amount of the bid in wei.
        target_calldata: Calldata for the contract call.
        chain_id: The chain ID to bid on.
        target_contract: The contract address to call.
        permission_key: The permission key to bid on.
        status: The latest status for bid.
        initiation_time: The time server received the bid formatted in rfc3339.
        profile_id: The profile id for the bid owner.
        gas_limit: The gas limit for the bid.
    """

    id: UUIDString
    bid_amount: IntString
    target_calldata: HexString
    chain_id: str
    target_contract: Address
    permission_key: HexString
    status: BidStatusEvm
    initiation_time: datetime
    profile_id: str | None = Field(default=None)
    gas_limit: IntString


class OpportunityDeleteEvm(BaseModel):
    """
    Attributes:
        chain_id: The chain ID for opportunities to be removed.
        permission_key: The permission key for the opportunities to be removed.
    """
    chain_id: str
    permission_key: HexString

    supported_versions: ClassVar[list[str]] = ["v1"]
    chain_type: ClassVar[str] = "evm"

    @model_validator(mode="before")
    @classmethod
    def check_version(cls, data):
        if data["version"] not in cls.supported_versions:
            raise UnsupportedOpportunityDeleteVersionException(
                f"Cannot handle opportunity version: {data['version']}. Please upgrade your client."
            )
        return data



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class GlobalConfigJSON(typing.TypedDict):
    emergency_mode: int
    flash_take_order_blocked: int
    new_orders_blocked: int
    orders_taking_blocked: int
    host_fee_bps: int
    is_order_taking_permissionless: int
    padding0: list[int]
    padding1: list[int]
    pda_authority_previous_lamports_balance: int
    total_tip_amount: int
    host_tip_amount: int
    pda_authority: str
    pda_authority_bump: int
    admin_authority: str
    padding2: list[int]


@dataclass
class GlobalConfig:
    discriminator: typing.ClassVar = b"\x95\x08\x9c\xca\xa0\xfc\xb0\xd9"
    layout: typing.ClassVar = borsh.CStruct(
        "emergency_mode" / borsh.U8,
        "flash_take_order_blocked" / borsh.U8,
        "new_orders_blocked" / borsh.U8,
        "orders_taking_blocked" / borsh.U8,
        "host_fee_bps" / borsh.U16,
        "is_order_taking_permissionless" / borsh.U8,
        "padding0" / borsh.U8[1],
        "padding1" / borsh.U64[10],
        "pda_authority_previous_lamports_balance" / borsh.U64,
        "total_tip_amount" / borsh.U64,
        "host_tip_amount" / borsh.U64,
        "pda_authority" / BorshPubkey,
        "pda_authority_bump" / borsh.U64,
        "admin_authority" / BorshPubkey,
        "padding2" / borsh.U64[247],
    )
    emergency_mode: int
    flash_take_order_blocked: int
    new_orders_blocked: int
    orders_taking_blocked: int
    host_fee_bps: int
    is_order_taking_permissionless: int
    padding0: list[int]
    padding1: list[int]
    pda_authority_previous_lamports_balance: int
    total_tip_amount: int
    host_tip_amount: int
    pda_authority: Pubkey
    pda_authority_bump: int
    admin_authority: Pubkey
    padding2: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["GlobalConfig"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["GlobalConfig"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["GlobalConfig"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "GlobalConfig":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = GlobalConfig.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            emergency_mode=dec.emergency_mode,
            flash_take_order_blocked=dec.flash_take_order_blocked,
            new_orders_blocked=dec.new_orders_blocked,
            orders_taking_blocked=dec.orders_taking_blocked,
            host_fee_bps=dec.host_fee_bps,
            is_order_taking_permissionless=dec.is_order_taking_permissionless,
            padding0=dec.padding0,
            padding1=dec.padding1,
            pda_authority_previous_lamports_balance=dec.pda_authority_previous_lamports_balance,
            total_tip_amount=dec.total_tip_amount,
            host_tip_amount=dec.host_tip_amount,
            pda_authority=dec.pda_authority,
            pda_authority_bump=dec.pda_authority_bump,
            admin_authority=dec.admin_authority,
            padding2=dec.padding2,
        )

    def to_json(self) -> GlobalConfigJSON:
        return {
            "emergency_mode": self.emergency_mode,
            "flash_take_order_blocked": self.flash_take_order_blocked,
            "new_orders_blocked": self.new_orders_blocked,
            "orders_taking_blocked": self.orders_taking_blocked,
            "host_fee_bps": self.host_fee_bps,
            "is_order_taking_permissionless": self.is_order_taking_permissionless,
            "padding0": self.padding0,
            "padding1": self.padding1,
            "pda_authority_previous_lamports_balance": self.pda_authority_previous_lamports_balance,
            "total_tip_amount": self.total_tip_amount,
            "host_tip_amount": self.host_tip_amount,
            "pda_authority": str(self.pda_authority),
            "pda_authority_bump": self.pda_authority_bump,
            "admin_authority": str(self.admin_authority),
            "padding2": self.padding2,
        }

    @classmethod
    def from_json(cls, obj: GlobalConfigJSON) -> "GlobalConfig":
        return cls(
            emergency_mode=obj["emergency_mode"],
            flash_take_order_blocked=obj["flash_take_order_blocked"],
            new_orders_blocked=obj["new_orders_blocked"],
            orders_taking_blocked=obj["orders_taking_blocked"],
            host_fee_bps=obj["host_fee_bps"],
            is_order_taking_permissionless=obj["is_order_taking_permissionless"],
            padding0=obj["padding0"],
            padding1=obj["padding1"],
            pda_authority_previous_lamports_balance=obj[
                "pda_authority_previous_lamports_balance"
            ],
            total_tip_amount=obj["total_tip_amount"],
            host_tip_amount=obj["host_tip_amount"],
            pda_authority=Pubkey.from_string(obj["pda_authority"]),
            pda_authority_bump=obj["pda_authority_bump"],
            admin_authority=Pubkey.from_string(obj["admin_authority"]),
            padding2=obj["padding2"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import EnumForCodegen, BorshPubkey
import borsh_construct as borsh

BoolJSONValue = tuple[bool]
U16JSONValue = tuple[int]
PubkeyJSONValue = tuple[str]
BoolValue = tuple[bool]
U16Value = tuple[int]
PubkeyValue = tuple[Pubkey]


class BoolJSON(typing.TypedDict):
    value: BoolJSONValue
    kind: typing.Literal["Bool"]


class U16JSON(typing.TypedDict):
    value: U16JSONValue
    kind: typing.Literal["U16"]


class PubkeyJSON(typing.TypedDict):
    value: PubkeyJSONValue
    kind: typing.Literal["Pubkey"]


@dataclass
class Bool:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Bool"
    value: BoolValue

    def to_json(self) -> BoolJSON:
        return BoolJSON(
            kind="Bool",
            value=(self.value[0],),
        )

    def to_encodable(self) -> dict:
        return {
            "Bool": {
                "item_0": self.value[0],
            },
        }


@dataclass
class U16:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "U16"
    value: U16Value

    def to_json(self) -> U16JSON:
        return U16JSON(
            kind="U16",
            value=(self.value[0],),
        )

    def to_encodable(self) -> dict:
        return {
            "U16": {
                "item_0": self.value[0],
            },
        }


@dataclass
class Pubkey:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Pubkey"
    value: PubkeyValue

    def to_json(self) -> PubkeyJSON:
        return PubkeyJSON(
            kind="Pubkey",
            value=(str(self.value[0]),),
        )

    def to_encodable(self) -> dict:
        return {
            "Pubkey": {
                "item_0": self.value[0],
            },
        }


UpdateGlobalConfigValueKind = typing.Union[Bool, U16, Pubkey]
UpdateGlobalConfigValueJSON = typing.Union[BoolJSON, U16JSON, PubkeyJSON]


def from_decoded(obj: dict) -> UpdateGlobalConfigValueKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Bool" in obj:
        val = obj["Bool"]
        return Bool((val["item_0"],))
    if "U16" in obj:
        val = obj["U16"]
        return U16((val["item_0"],))
    if "Pubkey" in obj:
        val = obj["Pubkey"]
        return Pubkey((val["item_0"],))
    raise ValueError("Invalid enum object")


def from_json(obj: UpdateGlobalConfigValueJSON) -> UpdateGlobalConfigValueKind:
    if obj["kind"] == "Bool":
        bool_json_value = typing.cast(BoolJSONValue, obj["value"])
        return Bool((bool_json_value[0],))
    if obj["kind"] == "U16":
        u16json_value = typing.cast(U16JSONValue, obj["value"])
        return U16((u16json_value[0],))
    if obj["kind"] == "Pubkey":
        pubkey_json_value = typing.cast(PubkeyJSONValue, obj["value"])
        return Pubkey((Pubkey.from_string(pubkey_json_value[0]),))
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Bool" / borsh.CStruct("item_0" / borsh.Bool),
    "U16" / borsh.CStruct("item_0" / borsh.U16),
    "Pubkey" / borsh.CStruct("item_0" / BorshPubkey),
)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class TakeOrderArgs(typing.TypedDict):
    input_amount: int
    min_output_amount: int
    tip_amount_permissionless_taking: int


layout = borsh.CStruct(
    "input_amount" / borsh.U64,
    "min_output_amount" / borsh.U64,
    "tip_amount_permissionless_taking" / borsh.U64,
)


class TakeOrderAccounts(typing.TypedDict):
    taker: Pubkey
    maker: Pubkey
    global_config: Pubkey
    pda_authority: Pubkey
    order: Pubkey
    input_mint: Pubkey
    output_mint: Pubkey
    input_vault: Pubkey
    taker_input_ata: Pubkey
    taker_output_ata: Pubkey
    maker_output_ata: Pubkey
    express_relay: Pubkey
    express_relay_metadata: Pubkey
    sysvar_instructions: Pubkey
    permission: typing.Optional[Pubkey]
    config_router: Pubkey
    input_token_program: Pubkey
    output_token_program: Pubkey


def take_order(
    args: TakeOrderArgs,
    accounts: TakeOrderAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["taker"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["maker"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["global_config"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["pda_authority"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["order"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["input_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["output_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["input_vault"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["taker_input_ata"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["taker_output_ata"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["maker_output_ata"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["express_relay"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["express_relay_metadata"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["sysvar_instructions"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["permission"], is_signer=False, is_writable=False)
        if accounts["permission"]
        else AccountMeta(pubkey=program_id, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["config_router"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["input_token_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["output_token_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xa3\xd0\x14\xac\xdfA\xff\xe4"
    encoded_args = layout.build(
        {
            "input_amount": args["input_amount"],
            "min_output_amount": args["min_output_amount"],
            "tip_amount_permissionless_taking": args[
                "tip_amount_permissionless_taking"
            ],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)
